# coding=utf-8
#
# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2020 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENCE.TXT
#
import time
import warnings
from typing import Optional, Tuple

try:
    import MySQLdb
except ImportError:
    MySQLdb = None

from clcommon.cpapi import db_access
from clcommon.cpapi.cpapiexceptions import NoDBAccessData, NotSupported
from clcommon.lib.config import Config
from cl_plus.consts import COLLECTORS_DEFAULT_INTERVAL

from .collector_base import CollectorBase


def _get_mysql_queries_num(dblogin: str, dbpass: str):
    """
    Retrieve MySQL queries (number)
    :param dblogin: admin login to DB
    :param dbpass: admin password to DB
    :return: (message, queries_num) tuple.
        message - 'OK' or error message
        queries_num - active MySQL queries number or 0 if DB error
    """
    if not MySQLdb:
        return "No MySQL client libraries installed", 0
    db = None
    try:
        db = MySQLdb.connect(host='localhost', user=dblogin, passwd=dbpass, db='mysql', use_unicode=True,
                             charset='utf8')
        cursor = db.cursor()
        # This query returns current running SQL queries
        # See https://stackoverflow.com/questions/7432241/mysql-show-status-active-or-total-connections
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=MySQLdb.Warning)
            query_prefix = 'SELECT query_id FROM information_schema.PROCESSLIST'
            cursor.execute(f'{query_prefix} WHERE command=\'Query\' AND info not LIKE \'{query_prefix}%\'')
        data = cursor.fetchall()
    except MySQLdb.MySQLError as e:
        return str(e), set()
    finally:
        if db is not None:
            db.close()
    unique_query_ids = set(row[0] for row in data)
    return 'OK', unique_query_ids


class MySQLCollector(CollectorBase):
    class DBAccess:
        def __init__(self, login: Optional[str] = None, password: Optional[str] = None):
            self.login = login
            self.password = password
            self.expire_time = time.time() + 3600

        def __eq__(self, other):
            return self.login == other.login and self.password == other.password

        def __bool__(self):
            return bool(self.login and self.password and not self.is_expired())

        def is_expired(self) -> bool:
            return time.time() > self.expire_time

    def __init__(self, _logger):
        super(MySQLCollector, self).__init__(_logger)
        self._is_mysql_error = False
        self.collection_interval = self.scrape_interval()
        self.access = self.DBAccess()

    def init(self):
        """
        Initialize MySQL collector
        :return: None
        """
        self._aggregated_data = set()
        self._logger.info("MySQL collector init")

    def scrape_interval(self):
        """
        Obtaining scrape interval for collector
         - if no config/no parameter in config -> use default
         - if value in config is invalid -> use default
         - if value in config is ok -> use it
        """
        interval = Config().get_value('mysql_collector_interval', COLLECTORS_DEFAULT_INTERVAL)
        try:
            interval = int(interval)
        except Exception:
            self._logger.warn('mysql_collector_interval parameter must be valid integer, applying default value')
            return COLLECTORS_DEFAULT_INTERVAL
        if interval > COLLECTORS_DEFAULT_INTERVAL or interval < 1:
            self._logger.warn('mysql_collector_interval parameter must be in range 1 to %s, applying default value',
                              str(COLLECTORS_DEFAULT_INTERVAL))
            return COLLECTORS_DEFAULT_INTERVAL
        self._logger.info('mysql_collector_interval parameter is %s, applying', str(interval))
        return interval

    def _get_db_access(self) -> Tuple[str, str]:
        """
        Get DB access data from cpapi function. Logs error
        :return: tuple (db_root_login, db_password)
            None, None if error
        """
        if not self.access:
            try:
                access = db_access()    # {'pass': str, 'login': str}
                self.access = self.DBAccess(access['login'], access['pass'])
            except (NoDBAccessData, NotSupported, KeyError, TypeError) as e:
                # Can't retrieve data
                if not self._is_mysql_error:
                    self._logger.warn("MySQL collector: can't obtain MySQL DB access: %s", str(e))
                    self._is_mysql_error = True
        return self.access.login, self.access.password

    def aggregate_new_data(self):
        """
        Retrieve and aggregate new data
        :return None
        """
        self._get_db_access()
        if not self.access:     # access data still is not set
            return
        # New data present - aggregate
        message, unique_query_ids = _get_mysql_queries_num(self.access.login, self.access.password)
        if message == 'OK':
            self._aggregated_data |= unique_query_ids
            self._is_mysql_error = False
            return
        # Retrieve data error
        if not self._is_mysql_error:
            self._logger.warn("MySQL collector can't obtain MySQL queries number: %s", message)
            self._is_mysql_error = True

    def get_averages(self):
        """
        Get collector's averages data
        :return: dict:
            { "mysql_queries_num": 16}
            or None if can't get data
        """
        mysql_queries_num = len(self._aggregated_data)
        self._aggregated_data = set()
        return {"mysql_queries_num": mysql_queries_num}
